@@ -1,19 +1,22 @@
 use std::fmt::Write;
 use std::ops::Range;
 use std::path::PathBuf;
+use std::sync::LazyLock;
 
 use ecow::eco_vec;
+use regex::{Captures, Regex};
 use tiny_skia as sk;
-use typst::diag::{SourceDiagnostic, Warned};
-use typst::html::HtmlDocument;
+use typst::diag::{SourceDiagnostic, SourceResult, Warned};
 use typst::layout::{Abs, Frame, FrameItem, PagedDocument, Transform};
 use typst::visualize::Color;
-use typst::{Document, WorldExt};
-use typst_pdf::PdfOptions;
+use typst::{World, WorldExt};
+use typst_html::HtmlDocument;
+use typst_pdf::{PdfOptions, PdfStandard, PdfStandards};
+use typst_syntax::{FileId, Lines};
 
-use crate::collect::{Attr, FileSize, NoteKind, Test};
+use crate::collect::{FileSize, NoteKind, Targets, Test};
 use crate::logger::TestResult;
-use crate::world::TestWorld;
+use crate::world::{TestWorld, system_path};
 
 /// Runs a single test.
 ///
@@ -63,14 +66,15 @@ impl<'a> Runner<'a> {
             log!(into: self.result.infos, "tree: {:#?}", self.test.source.root());
         }
 
-        let html = self.test.attrs.contains(&Attr::Html);
-        let render = !html || self.test.attrs.contains(&Attr::Render);
-        if render {
+        if self.test.attrs.targets.contains(Targets::RENDER) {
             self.run_test::<PagedDocument>();
         }
-        if html {
+        if self.test.attrs.targets.contains(Targets::HTML) {
             self.run_test::<HtmlDocument>();
         }
+        if self.test.attrs.targets.contains(Targets::PDFTAGS) {
+            self.run_test::<Pdftags>();
+        }
 
         self.handle_not_emitted();
         self.handle_not_annotated();
@@ -80,18 +84,27 @@ impl<'a> Runner<'a> {
 
     /// Run test specific to document format.
     fn run_test<D: OutputType>(&mut self) {
-        let Warned { output, warnings } = typst::compile(&self.world);
-        let (doc, errors) = match output {
+        let Warned { output, warnings } = D::compile(&self.world, self.test);
+        let (doc, mut errors) = match output {
             Ok(doc) => (Some(doc), eco_vec![]),
             Err(errors) => (None, errors),
         };
 
-        if doc.is_none() && errors.is_empty() {
+        D::check_custom(self, doc.as_ref());
+
+        let output = doc.and_then(|doc| match doc.make_live() {
+            Ok(live) => Some((doc, live)),
+            Err(list) => {
+                errors.extend(list);
+                None
+            }
+        });
+
+        if output.is_none() && errors.is_empty() {
             log!(self, "no document, but also no errors");
         }
 
-        D::check_custom(self, doc.as_ref());
-        self.check_output(doc.as_ref());
+        self.check_output(output);
 
         for error in &errors {
             self.check_diagnostic(NoteKind::Error, error);
@@ -117,7 +130,7 @@ impl<'a> Runner<'a> {
             if seen {
                 continue;
             }
-            let note_range = self.format_range(&note.range);
+            let note_range = self.format_range(note.file, &note.range);
             if first {
                 log!(self, "not emitted");
                 first = false;
@@ -127,12 +140,12 @@ impl<'a> Runner<'a> {
     }
 
     /// Check that the document output is correct.
-    fn check_output<D: OutputType>(&mut self, document: Option<&D>) {
+    fn check_output<D: OutputType>(&mut self, output: Option<(D, D::Live)>) {
         let live_path = D::live_path(&self.test.name);
         let ref_path = D::ref_path(&self.test.name);
         let ref_data = std::fs::read(&ref_path);
 
-        let Some(document) = document else {
+        let Some((document, live)) = output else {
             if ref_data.is_ok() {
                 log!(self, "missing document");
                 log!(self, "  ref       | {}", ref_path.display());
@@ -140,7 +153,7 @@ impl<'a> Runner<'a> {
             return;
         };
 
-        let skippable = match D::is_skippable(document) {
+        let skippable = match D::is_skippable(&document) {
             Ok(skippable) => skippable,
             Err(()) => {
                 log!(self, "document has zero pages");
@@ -156,12 +169,22 @@ impl<'a> Runner<'a> {
         }
 
         // Render and save live version.
-        let live = document.make_live();
-        document.save_live(&self.test.name, &live);
+        if let Err(errors) = document.save_live(&self.test.name, &live) {
+            log!(self, "failed to save live version:");
+            for e in errors {
+                if let Some(file) = e.span.id() {
+                    let range = self.world.range(e.span);
+                    let diag_range = self.format_range(file, &range);
+                    log!(self, "  Error: {diag_range} {}", e.message);
+                } else {
+                    log!(self, "  Error: {}", e.message);
+                }
+            }
+        }
 
         // Compare against reference output if available.
         // Test that is ok doesn't need to be updated.
-        if ref_data.as_ref().map_or(false, |r| D::matches(&live, r)) {
+        if ref_data.as_ref().is_ok_and(|r| D::matches(&live, r)) {
             return;
         }
 
@@ -174,14 +197,18 @@ impl<'a> Runner<'a> {
                 );
             } else {
                 let ref_data = D::make_ref(live);
-                if !self.test.attrs.contains(&Attr::Large)
-                    && ref_data.len() > crate::REF_LIMIT
-                {
+                if !self.test.attrs.large && ref_data.len() > crate::REF_LIMIT {
                     log!(self, "reference output would exceed maximum size");
                     log!(self, "  maximum   | {}", FileSize(crate::REF_LIMIT));
                     log!(self, "  size      | {}", FileSize(ref_data.len()));
-                    log!(self, "please try to minimize the size of the test (smaller pages, less text, etc.)");
-                    log!(self, "if you think the test cannot be reasonably minimized, mark it as `large`");
+                    log!(
+                        self,
+                        "please try to minimize the size of the test (smaller pages, less text, etc.)"
+                    );
+                    log!(
+                        self,
+                        "if you think the test cannot be reasonably minimized, mark it as `large`"
+                    );
                     return;
                 }
                 std::fs::write(&ref_path, &ref_data).unwrap();
@@ -208,22 +235,17 @@ impl<'a> Runner<'a> {
     /// Compare a subset of notes with a given kind against diagnostics of
     /// that same kind.
     fn check_diagnostic(&mut self, kind: NoteKind, diag: &SourceDiagnostic) {
-        // Ignore diagnostics from other sources than the test file itself.
-        if diag.span.id().is_some_and(|id| id != self.test.source.id()) {
-            return;
-        }
         // TODO: remove this once HTML export is stable
         if diag.message == "html export is under active development and incomplete" {
             return;
         }
 
-        let message = diag.message.replace("\\", "/");
         let range = self.world.range(diag.span);
-        self.validate_note(kind, range.clone(), &message);
+        self.validate_note(kind, diag.span.id(), range.clone(), &diag.message);
 
         // Check hints.
         for hint in &diag.hints {
-            self.validate_note(NoteKind::Hint, range.clone(), hint);
+            self.validate_note(NoteKind::Hint, diag.span.id(), range.clone(), hint);
         }
     }
 
@@ -235,15 +257,27 @@ impl<'a> Runner<'a> {
     fn validate_note(
         &mut self,
         kind: NoteKind,
+        file: Option<FileId>,
         range: Option<Range<usize>>,
         message: &str,
     ) {
+        // HACK: Replace backslashes path sepators with slashes for cross
+        // platform reproducible error messages.
+        static RE: LazyLock<Regex> =
+            LazyLock::new(|| Regex::new("\\((.*) (at|in) (.+)\\)").unwrap());
+        let message = RE.replace(message, |caps: &Captures| {
+            let path = caps[3].replace('\\', "/");
+            format!("({} {} {})", &caps[1], &caps[2], path)
+        });
+
         // Try to find perfect match.
+        let file = file.unwrap_or(self.test.source.id());
         if let Some((i, _)) = self.test.notes.iter().enumerate().find(|&(i, note)| {
             !self.seen[i]
                 && note.kind == kind
                 && note.range == range
                 && note.message == message
+                && note.file == file
         }) {
             self.seen[i] = true;
             return;
@@ -257,7 +291,7 @@ impl<'a> Runner<'a> {
                 && (note.range == range || note.message == message)
         }) else {
             // Not even a close match, diagnostic is not annotated.
-            let diag_range = self.format_range(&range);
+            let diag_range = self.format_range(file, &range);
             log!(into: self.not_annotated, "  {kind}: {diag_range} {}", message);
             return;
         };
@@ -267,10 +301,10 @@ impl<'a> Runner<'a> {
 
         // Range is wrong.
         if range != note.range {
-            let note_range = self.format_range(&note.range);
-            let note_text = self.text_for_range(&note.range);
-            let diag_range = self.format_range(&range);
-            let diag_text = self.text_for_range(&range);
+            let note_range = self.format_range(note.file, &note.range);
+            let note_text = self.text_for_range(note.file, &note.range);
+            let diag_range = self.format_range(file, &range);
+            let diag_text = self.text_for_range(file, &range);
             log!(self, "mismatched range ({}):", note.pos);
             log!(self, "  message   | {}", note.message);
             log!(self, "  annotated | {note_range:<9} | {note_text}");
@@ -286,51 +320,69 @@ impl<'a> Runner<'a> {
     }
 
     /// Display the text for a range.
-    fn text_for_range(&self, range: &Option<Range<usize>>) -> String {
+    fn text_for_range(&self, file: FileId, range: &Option<Range<usize>>) -> String {
         let Some(range) = range else { return "No text".into() };
         if range.is_empty() {
-            "(empty)".into()
-        } else {
-            format!("`{}`", self.test.source.text()[range.clone()].replace('\n', "\\n"))
+            return "(empty)".into();
         }
+
+        let lines = self.lookup(file);
+        lines.text()[range.clone()].replace('\n', "\\n").replace('\r', "\\r")
     }
 
     /// Display a byte range as a line:column range.
-    fn format_range(&self, range: &Option<Range<usize>>) -> String {
+    fn format_range(&self, file: FileId, range: &Option<Range<usize>>) -> String {
         let Some(range) = range else { return "No range".into() };
+
+        let mut preamble = String::new();
+        if file != self.test.source.id() {
+            preamble = format!("\"{}\" ", system_path(file).unwrap().display());
+        }
+
         if range.start == range.end {
-            self.format_pos(range.start)
+            format!("{preamble}{}", self.format_pos(file, range.start))
         } else {
-            format!("{}-{}", self.format_pos(range.start,), self.format_pos(range.end,))
+            format!(
+                "{preamble}{}-{}",
+                self.format_pos(file, range.start),
+                self.format_pos(file, range.end)
+            )
         }
     }
 
     /// Display a position as a line:column pair.
-    fn format_pos(&self, pos: usize) -> String {
-        if let (Some(line_idx), Some(column_idx)) =
-            (self.test.source.byte_to_line(pos), self.test.source.byte_to_column(pos))
-        {
-            let line = self.test.pos.line + line_idx;
-            let column = column_idx + 1;
-            if line == 1 {
-                format!("{column}")
-            } else {
-                format!("{line}:{column}")
-            }
+    fn format_pos(&self, file: FileId, pos: usize) -> String {
+        let lines = self.lookup(file);
+
+        let res = lines.byte_to_line_column(pos).map(|(line, col)| (line + 1, col + 1));
+        let Some((line, col)) = res else {
+            return "oob".into();
+        };
+
+        if line == 1 { format!("{col}") } else { format!("{line}:{col}") }
+    }
+
+    #[track_caller]
+    fn lookup(&self, file: FileId) -> Lines<String> {
+        if self.test.source.id() == file {
+            self.test.source.lines().clone()
         } else {
-            "oob".into()
+            self.world.lookup(file)
         }
     }
 }
 
 /// An output type we can test.
-trait OutputType: Document {
+trait OutputType: Sized {
     /// The type that represents live output.
     type Live;
 
     /// The path at which the live output is stored.
     fn live_path(name: &str) -> PathBuf;
 
+    /// Compiles the Typst test into this output type.
+    fn compile(world: &dyn World, test: &Test) -> Warned<SourceResult<Self>>;
+
     /// The path at which the reference output is stored.
     fn ref_path(name: &str) -> PathBuf;
 
@@ -340,10 +392,10 @@ trait OutputType: Document {
     }
 
     /// Produces the live output.
-    fn make_live(&self) -> Self::Live;
+    fn make_live(&self) -> SourceResult<Self::Live>;
 
     /// Saves the live output.
-    fn save_live(&self, name: &str, live: &Self::Live);
+    fn save_live(&self, name: &str, live: &Self::Live) -> SourceResult<()>;
 
     /// Produces the reference output from the live output.
     fn make_ref(live: Self::Live) -> Vec<u8>;
@@ -367,6 +419,10 @@ impl OutputType for PagedDocument {
         format!("{}/{}.png", crate::REF_PATH, name).into()
     }
 
+    fn compile(world: &dyn World, _: &Test) -> Warned<SourceResult<Self>> {
+        typst::compile(world)
+    }
+
     fn is_skippable(&self) -> Result<bool, ()> {
         /// Whether rendering of a frame can be skipped.
         fn skippable_frame(frame: &Frame) -> bool {
@@ -387,11 +443,11 @@ impl OutputType for PagedDocument {
         }
     }
 
-    fn make_live(&self) -> Self::Live {
-        render(self, 1.0)
+    fn make_live(&self) -> SourceResult<Self::Live> {
+        Ok(render(self, 1.0))
     }
 
-    fn save_live(&self, name: &str, live: &Self::Live) {
+    fn save_live(&self, name: &str, live: &Self::Live) -> SourceResult<()> {
         // Save live version, possibly rerendering if different scale is
         // requested.
         let mut pixmap_live = live;
@@ -407,7 +463,7 @@ impl OutputType for PagedDocument {
         // Write PDF if requested.
         if crate::ARGS.pdf() {
             let pdf_path = format!("{}/pdf/{}.pdf", crate::STORE_PATH, name);
-            let pdf = typst_pdf::pdf(self, &PdfOptions::default()).unwrap();
+            let pdf = typst_pdf::pdf(self, &PdfOptions::default())?;
             std::fs::write(pdf_path, pdf).unwrap();
         }
 
@@ -417,6 +473,8 @@ impl OutputType for PagedDocument {
             let svg = typst_svg::svg_merged(self, Abs::pt(5.0));
             std::fs::write(svg_path, svg).unwrap();
         }
+
+        Ok(())
     }
 
     fn make_ref(live: Self::Live) -> Vec<u8> {
@@ -452,13 +510,17 @@ impl OutputType for HtmlDocument {
         format!("{}/html/{}.html", crate::REF_PATH, name).into()
     }
 
-    fn make_live(&self) -> Self::Live {
-        // TODO: Do this earlier to be able to process export errors.
-        typst_html::html(self).unwrap()
+    fn compile(world: &dyn World, _: &Test) -> Warned<SourceResult<Self>> {
+        typst::compile(world)
     }
 
-    fn save_live(&self, name: &str, live: &Self::Live) {
+    fn make_live(&self) -> SourceResult<Self::Live> {
+        typst_html::html(self)
+    }
+
+    fn save_live(&self, name: &str, live: &Self::Live) -> SourceResult<()> {
         std::fs::write(Self::live_path(name), live).unwrap();
+        Ok(())
     }
 
     fn make_ref(live: Self::Live) -> Vec<u8> {
@@ -470,6 +532,64 @@ impl OutputType for HtmlDocument {
     }
 }
 
+struct Pdftags(String);
+
+impl OutputType for Pdftags {
+    type Live = String;
+
+    fn live_path(name: &str) -> PathBuf {
+        format!("{}/pdftags/{}.yml", crate::STORE_PATH, name).into()
+    }
+
+    fn ref_path(name: &str) -> PathBuf {
+        format!("{}/pdftags/{}.yml", crate::REF_PATH, name).into()
+    }
+
+    fn compile(world: &dyn World, test: &Test) -> Warned<SourceResult<Self>> {
+        let Warned { output, warnings } = typst::compile::<PagedDocument>(world);
+        let mut doc = match output {
+            Ok(doc) => doc,
+            Err(errors) => return Warned { output: Err(errors), warnings },
+        };
+        if doc.info.title.is_none() {
+            doc.info.title = Some("<test>".into());
+        }
+
+        let standards = if test.attrs.pdf_ua {
+            PdfStandards::new(&[PdfStandard::Ua_1]).unwrap()
+        } else {
+            PdfStandards::default()
+        };
+        let options = PdfOptions { standards, ..Default::default() };
+        let output = typst_pdf::pdf_tags(&doc, &options).map(Pdftags);
+        Warned { warnings, output }
+    }
+
+    fn is_skippable(&self) -> Result<bool, ()> {
+        Ok(self.0.is_empty())
+    }
+
+    fn make_live(&self) -> SourceResult<Self::Live> {
+        Ok(self.0.clone())
+    }
+
+    fn save_live(&self, name: &str, live: &Self::Live) -> SourceResult<()> {
+        std::fs::write(Self::live_path(name), live).unwrap();
+        Ok(())
+    }
+
+    fn make_ref(live: Self::Live) -> Vec<u8> {
+        live.into_bytes()
+    }
+
+    fn matches(live: &Self::Live, ref_data: &[u8]) -> bool {
+        // Compare lines with newlines stripped, since on windows git might
+        // check out the ref files with CRLF line endings.
+        let ref_str = std::str::from_utf8(ref_data).unwrap();
+        ref_str.lines().eq(live.lines())
+    }
+}
+
 /// Draw all frames into one image with padding in between.
 fn render(document: &PagedDocument, pixel_per_pt: f32) -> sk::Pixmap {
     for page in &document.pages {
